#!/usr/bin/python
#
# Copyright 2005 Lars Wirzenius (liw@iki.fi)
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


"""Debian package installation and uninstallation tester.

This program sets up a minimal Debian system in a chroot, and installs
and uninstalls packages and their dependencies therein, looking for
problems.

See the manual page (dit.1, generated from dit.docbook) for more usage
information.

Lars Wirzenius <liw@iki.fi>
"""


VERSION = "0.4"


import time
import getopt
import sys
import commands
import tempfile
import shutil
import os
import tarfile
import stat


class Settings:

    """Global settings for this program."""
    
    def __init__(self):
        self.args_are_package_files = True
        self.really_verbose = False
        self.verbose = False
        self.debian_mirror = "http://ftp.debian.org/"
        self.debian_distro = "unstable"
        self.basetgz = None
        self.savetgz = None
        self.log_files = [sys.stdout]
        self.ignored_files = [
            "/etc/ld.so.cache",
            "/usr/share/info/dir",
            "/usr/share/info/dir.old",
            "/var/backups/infodir.bak",
            "/var/cache/apt/archives/lock",
            "/var/cache/apt/pkgcache.bin", 
            "/var/cache/debconf",
            "/var/cache/debconf/config.dat",
            "/var/cache/debconf/config.dat-old",
            "/var/cache/debconf/passwords.dat",
            "/var/cache/debconf/templates.dat",
            "/var/cache/debconf/templates.dat-old",
            "/var/lib/dpkg/available", 
            "/var/lib/dpkg/available-old", 
            "/var/lib/dpkg/lock", 
            "/var/lib/dpkg/status", 
            "/var/lib/dpkg/status-old", 
            "/",
            ]


settings = Settings()
startup_time = time.time()


def log_msg(log_to_default, timestamp, prefix, msg, exit_code):
    """Write log message to stdout.
        'log_to_default' tells whether to log to stdout.
        'timestamp' is None or the the time (from time.time()).
       'prefix' is "DEBUG", or other prefix to add to message.
       'msg' is the actual message.
       If 'exit_code' is non-zero, exit process with that exit code."""
    if timestamp:
        t = timestamp - startup_time
        min = int(t) / 60
        sec = t % 60
    for f in settings.log_files:
        if f == sys.stdout and not log_to_default:
            continue
        if timestamp:
            f.write("%dmin%.1fs " % (min, sec))
        if prefix:
            f.write("%s: " % prefix)
        f.write(msg)
        f.write("\n")
        f.flush()
    if exit_code:
        sys.exit(exit_code)


def dump(msg):
    """Write really verbose debugging message."""
    log_msg(settings.really_verbose, None, "", msg, 0)


def debug(msg):
    """Write a debugging log message."""
    log_msg(settings.verbose, time.time(), "DEBUG", msg, 0)


def info(msg):
    """Write an informational log message."""
    log_msg(True, time.time(), "", msg, 0)


def error(msg, exit_code=1):
    """Write an error message and exit (unless exit_code is 0)."""
    log_msg(True, time.time(), "ERROR", msg, exit_code=exit_code)


def indent_string(str):
    """Indent all lines in a string with two spaces and return result."""
    return "\n".join(["  " + line for line in str.split("\n")])


def run(command, ignore_errors=False):
    """Run an external command and die with error message if it failes."""
    debug("Starting command: %s" % command)
    f = os.popen("{ export LC_ALL=C LANGUAGES=; %s ; } 2>&1" % 
                 command, "r")
    output = ""
    while 1:
        line = f.readline()
        if not line:
            break
        output += line
        if line[-1:] == "\n":
            line = line[:-1]
        dump("  " + line)
    status = f.close()
    if status is None:
        status = 0
    failed = not os.WIFEXITED(status) or os.WEXITSTATUS(status) != 0
    if not failed:
        debug("Command ok: %s" % repr(command))
    elif ignore_errors:
        debug("Command failed (status=%d), but ignoring error: %s" % 
              (status, repr(command)))
    else:
        error("Command failed (status=%d): %s" % (status, repr(command)))
    return status, output


def run_in_chroot(root, command, ignore_errors=False):
    return run("chroot %s %s" % (root, command), ignore_errors=ignore_errors)


def create_temp_dir():
    """Create a temporary directory and return its full path."""
    path = tempfile.mkdtemp(dir=".")
    os.chmod(path, 0755)
    debug("Created temporary directory %s" % path)
    return path


def create_temp_file():
    """Create a temporary file and return its full path."""
    (fd, path) = tempfile.mkstemp(dir=".")
    debug("Created temporary file %s" % path)
    return (fd, path)


def remove_dir_tree(path):
    """Remove a directory tree, including all files and contents."""
    shutil.rmtree(path)
    debug("Removed directory tree at %s" % path)


def create_file(name, contents):
    """Create a new file with the desired name and contents."""
    try:
        f = file(name, "w")
        f.write(contents)
        f.close()
    except IOError, detail:
        error("Couldn't create file %s: %s" % (name, detail))


def copy_files(source_names, target_name):
    """Copy files in 'source_name' to file/dir 'target_name'."""
    debug("Copying %s to %s" % (", ".join(source_names), target_name))
    for source_name in source_names:
        try:
            shutil.copy(source_name, target_name)
        except IOError, detail:
            error("Error copying %s to %s: %s" % 
                  (source_name, target_name, detail))


def remove_files(filenames):
    """Remove some files."""
    for filename in filenames:
        debug("Removing %s" % filename)
        try:
            os.remove(filename)
        except OSError, detail:
            error("Couldn't remove %s: %s" % (filename, detail))


def setup_minimal_chroot(path):
    """Set up a minimal Debian system in a chroot at 'path'."""
    debug("Setting up minimal chroot for %s at %s." % 
          (settings.debian_distro, path))
    run("debootstrap %s '%s' '%s'" % 
        (settings.debian_distro, path, settings.debian_mirror))
    create_file(os.path.join(path, "etc/apt/sources.list"),
                "deb %s %s main\n" % 
                    (settings.debian_mirror, settings.debian_distro))


def save_dir_tree(path, result):
    """Tar and compress all files in the directory tree at 'path'."""
    debug("Saving %s to %s." % (path, result))
    try:
        tf = tarfile.open(result, "w:gz")
        tf.add(path, arcname=".")
        tf.close()
    except tarfile.TarError, detail:
        error("Couldn't create tar file %s: %s" % (result, detail))


def unpack_tarball_to_dir(tarball, path):
    """Unpack a tarball to a directory (cf. 'tar -C path -xzf tarball')."""
    debug("Unpacking %s into %s" % (tarball, path))
    run("tar -C '%s' -zxf '%s'" % (path, tarball))

    return
    try:
        tf = tarfile.open(tarball, "r|gz")
        for ti in tf:
            tf.extract(ti, path)
        tf.close()
    except tarfile.TarError, detail:
        error("Couldn't extract tar file %s into %s: %s" % 
              (tarball, path, detail))


def save_dir_tree_meta_data(root):
    """Return the meta data for all objects in a dir tree."""
    root = os.path.join(root, ".")
    dict = {}
    for dirpath, dirnames, filenames in os.walk(root):
        assert dirpath[:len(root)] == root
        for name in [dirpath] + [os.path.join(dirpath, f) for f in filenames]:
            dict[name[len(root):]] = os.lstat(name)
    return dict


def objects_are_different(m1, m2):
    """Are filesystem objects different based on their meta data?"""
    if (m1.st_mode != m2.st_mode or 
        m1.st_uid != m2.st_uid or m1.st_gid != m2.st_gid):
        return True
    if stat.S_ISREG(m1.st_mode):
        return m1.st_size != m2.st_size or m1.st_mtime != m2.st_mtime
# We don't have access to the linkname yet. XXX
#    if stat.S_ISLNK(m1.st_mode):
#        return m1.linkname != m2.linkname
    return False


def diff_meta_data(tree1, tree2):
    """Compare two dir trees and return list of new files (only in 'tree2'),
       removed files (only in 'tree1'), and modified files."""

    tree1 = tree1.copy()
    tree2 = tree2.copy()

    for name in settings.ignored_files:
        if name in tree1:
            del tree1[name]
        if name in tree2:
            del tree2[name]

    modified = []
    for name in tree1.keys()[:]:
        if name in tree2:
            if objects_are_different(tree1[name], tree2[name]):
                modified.append((name, tree1[name]))
            del tree1[name]
            del tree2[name]

    removed = [x for x in tree1.iteritems()]
    new = [x for x in tree2.iteritems()]

    return new, removed, modified


def file_list(meta_infos):
    """Return list of indented filenames."""
    return "\n".join(["  " + name for (name, data) in meta_infos])


def get_chroot_selections(root):
    """Get current package selections in a chroot."""
    (status, output) = run_in_chroot(root, "dpkg --get-selections '*'")
    list = [line.split() for line in output.split("\n") if line.strip()]
    dict = {}
    for name, status in list:
        dict[name] = status
    return dict


def diff_selections(root, selections):
    """Compare original and current package selection.
       Return dict where dict[package_name] = original_status, that is,
       the value in the dict is the state that the package needs to be
       set to to restore original selections."""
    changes = {}
    current = get_chroot_selections(root)
    for name, value in current.iteritems():
        if name not in selections:
            changes[name] = "purge"
        elif selections[name] != current[name] and \
             selections[name] == "purge":
            changes[name] = selections[name]
    return changes


def restore_chroot_selections(root, changes):
    """Restore package selections in a chroot by applying 'changes'.
       'changes' is a return value from diff_selections."""

    # We first remove all packages that need to be purged. This is so that
    # we can test that the postrm scripts only use essential stuff.
    for name, state in changes.iteritems():
        if state in ["purge", "remove"]:
            run_in_chroot(root, "dpkg --remove %s" % name, ignore_errors=True)
    run_in_chroot(root, "dpkg --remove --pending", ignore_errors=True)
    for name, state in changes.iteritems():
        if state == "purge":
            run_in_chroot(root, "dpkg --remove %s" % name, ignore_errors=True)
    run_in_chroot(root, "dpkg --purge --pending")


def get_package_names_from_package_files(filenames):
    """Return list of package names given list of package file names."""
    list = []
    for filename in filenames:
        (status, output) = run("dpkg --info " + filename)
        for line in [line.lstrip() for line in output.split("\n")]:
            if line[:len("Package:")] == "Package:":
                list.append(line.split(":", 1)[1].strip())
    return list


def apt_get_knows(root, package_names):
    """Does apt-get (or apt-cache) in chroot know about a set of packages?"""
    for name in package_names:
        (status, output) = run_in_chroot(root, "apt-cache show " + name,
                                             ignore_errors=True)
        if not os.WIFEXITED(status):
            error("Error occurred when running apt-cache in chroot:\n" +
                  output)
        if os.WEXITSTATUS(status) != 0 or not output.strip():
            return False
    return True
        


def parse_command_line():
    """Parse the command line, change global settings, return non-options."""
    (opts, args) = getopt.getopt(sys.argv[1:], "ab:d:hi:l:m:ps:vV",
                                 ["distribution=", "mirror=", "verbose",
                                  "log-file=", "ignore=", "help",
                                  "basetgz=", "save=", "version",
                                  "apt", "--pbuilder"])
    exit = None
    for opt, optarg in opts:
        if opt in ["-b", "--basetgz"]:
            settings.basetgz = optarg
        elif opt in ["-p", "--pbuilder"]:
            settings.basetgz = "/var/cache/pbuilder/base.tgz"
        elif opt in ["-s", "--save"]:
            settings.savetgz = optarg
        elif opt in ["-d", "--distribution"]:
            settings.debian_distro = optarg
        elif opt in ["-a", "--apt"]:
            settings.args_are_package_files = False
        elif opt in ["-h", "--help"]:
            sys.stdout.write("Usage: dit [options] package-file ...\n")
            sys.stdout.write("See manual page dit(1) for more help.\n")
            exit = 0
        elif opt in ["-V", "--version"]:
            sys.stdout.write("Version: %s\n" % VERSION)
            exit = 0
        elif opt in ["-i", "--ignore"]:
            settings.ignored_files.append(optarg)
        elif opt in ["-m", "--mirror"]:
            settings.debian_mirror = optarg
        elif opt in ["-v", "--verbose"]:
            if not settings.verbose:
                settings.verbose = True
                settings.really_verbose = False
            else:
                settings.verbose = True
                settings.really_verbose = True
        elif opt in ["-l", "--log-file"]:
            try:
                f = file(optarg, "a")
            except IOError, detail:
                error("Can't open log file %s: %s" % (optarg, detail))
                exit = 1
            settings.log_files.append(f)

    if exit is not None:
        sys.exit(exit)

    return args
    

def main():
    """Main program."""

    args = parse_command_line()

    # Make sure debconf does not ask questions and stop everything.
    # Packages that don't use debconf will lose.
    os.environ["DEBIAN_FRONTEND"] = "noninteractive"

    # Create the chroot, either from an existing tarball (fast) or by
    # running debootstrap and stuff (slow).
    root = create_temp_dir()
    if settings.basetgz:
        unpack_tarball_to_dir(settings.basetgz, root)
        run_in_chroot(root, "apt-get update")
        run_in_chroot(root, "apt-get clean")
    else:
        setup_minimal_chroot(root)
        run_in_chroot(root, "apt-get update")
        run_in_chroot(root, "apt-get clean")
        if settings.savetgz:
            save_dir_tree(root, settings.savetgz)

    # Remember initial meta data and package selections.
    root_info = save_dir_tree_meta_data(root)
    selections = get_chroot_selections(root)

    # Install packages into the chroot.
    if settings.args_are_package_files:
        root_tmp = os.path.join(root, "tmp")
        copy_files(args, root_tmp)
        tmp_files = [os.path.basename(a) for a in args]
        tmp_files = [os.path.join("tmp", name) for name in tmp_files]
        run_in_chroot(root, "dpkg -i " + " ".join(tmp_files), 
                      ignore_errors=True)
        run_in_chroot(root, "apt-get -yf install")
        run_in_chroot(root, "apt-get clean")
        remove_files([os.path.join(root, name) for name in tmp_files])
    else:
        run_in_chroot(root, "apt-get -y install " +
                            " ".join(["'%s'" % arg for arg in args]))
        run_in_chroot(root, "apt-get clean")

    # Remove all packages from the chroot that weren't there initially.    
    changes = diff_selections(root, selections)
    restore_chroot_selections(root, changes)

    # Get meta data at the end.
    purged_info = save_dir_tree_meta_data(root)

    # Find and report problems.
    (new, removed, modified) = diff_meta_data(root_info, purged_info)
    if new:
        error("Package purging left files on system:\n" + file_list(new), 0)
    if removed:
        error("After purging files have disappeared:\n" +
              file_list(removed), 0)
    if modified:
        error("After purging files have been modified:\n" +
              file_list(modified), 0)

    # If there were any errors, quit now. Otherwise, report success.
    if new or removed or modified:
        remove_dir_tree(root)
        error("Installation and purging test FAILED.")
    info("Installation and purging test PASSED.")

    # All went well. Let's try an upgrade test. Install what we assume is
    # the previous version via apt-get, then the package files again.
    # Obviously none of this works if the arguments are package names or
    # if the packages aren't available via apt-get at all.
    if not settings.args_are_package_files:
        info("Can't test upgrades: -a or --apt option used.")
    else:
        package_names = get_package_names_from_package_files(args)
        debug("Package names: " + ", ".join(package_names))
        if not apt_get_knows(root, package_names):
            info("Can't test upgrade: packages not known by apt-get.")
        else:
            # First install via apt-get.
            run_in_chroot(root, 
                          "apt-get -yf install " + " ".join(package_names))
    
            # Then from the package files.
            copy_files(args, root_tmp)
            run_in_chroot(root, "dpkg -i " + " ".join(tmp_files), 
                          ignore_errors=True)
            run_in_chroot(root, "apt-get -yf install")
            run_in_chroot(root, "apt-get clean")
            remove_files([os.path.join(root, name) for name in tmp_files])
    
            # Remove all packages from the chroot that weren't there
            # initially.
            changes = diff_selections(root, selections)
            restore_chroot_selections(root, changes)
            
            # Get meta data at the end.
            purged_info = save_dir_tree_meta_data(root)
        
            # Find and report problems.
            (new, removed, modified) = diff_meta_data(root_info, purged_info)
            if new:
                error("Package purging left files on system:\n" + 
                        file_list(new), 
                      0)
            if removed:
                error("After purging files have disappeared:\n" +
                      file_list(removed), 0)
            if modified:
                error("After purging files have been modified:\n" +
                      file_list(modified), 0)
        
            # If there were any errors, quit now.
            if new or removed or modified:
                remove_dir_tree(root)
                error("Upgrade and purging test FAILED.")
    
            remove_dir_tree(root)
            info("Installation, upgrade and purging tests PASSED.")


if __name__ == "__main__":
    main()
